/**
 * Ein Labyrinth Spiel, bei eine Spielfigur durch ein Labyrinth bewegt
 * wird, bis alle Schaetze gefunden wurden.
 */
public class Labyrinth {

    private int breite = 19;

    /**
     * @return Die Breite des Labyrinths.
     */
    public int getBreite() {
        return breite;
    }

    /**
     * Setzt die Breite des Labyrinths, die im naechsten Spiel
     * verwendet wird.
     * <p>
     * Darf nur vor dem Spielbegin verwendet werden.
     * </p>
     */
    public void setBreite(int _breite) {
        breite = _breite;
    }

    private int hoehe = 19;

    /**
     * @return Die Hoehe des Labyrinths.
     */
    public int getHoehe() {
        return hoehe;
    }

    /**
     * Setzt die Hoehe des Labyrinths, die im naechsten Spiel
     * verwendet wird.
     * <p>
     * Darf nur vor dem Spielbegin verwendet werden.
     * </p>
     */
    public void setHoehe(int _hoehe) {
        hoehe = _hoehe;
    }

    private int schaetze = 5;

    /**
     * @return Die Zahl der im Labyrinth verbleibenden Schaetze.
     */
    public int getSchaetze() {
        return schaetze;
    }

    /**
     * Aendert die Zahl der Schaetze.
     * <p>
     * Darf nur vor dem Spielbegin verwendet werden.
     * </p>
     */
    public void setSchaetze(int _schaetze) {
        schaetze = _schaetze;
    }

    private int spielerZeile = 0;
    private int spielerSpalte = 1;

    private LabyrinthFeld[][] spielfeld;

    /**
     * Initialisiert das Attribut {@code spielfeld} mit einem
     * zufaellig generierten Labyrinth.
     * <p>
     * Hierbei wird die per set-Methoden gesetzten Parameter {@code
     * breite}, {@code hoehe} und {@code schaetze} beachtet.
     * </p>
     */
    public void labyrinthErstellen() {
        spielfeld =
            LabyrinthGenerator.generateLabyrinth(
                getBreite(), getHoehe(),
                getSchaetze());
    }

    /**
     * Fuehrt das Labyrinth Spiel aus.
     * @param args Wird ignoriert.
     */
    public static void main(String[] args) {
        Labyrinth l = new Labyrinth();

        System.out.print("Breite des Labyrinths: ");
        l.setBreite(
            Integer.parseInt(System.console().readLine()));

        System.out.print("Hoehe des Labyrinths: ");
        l.setHoehe(
            Integer.parseInt(System.console().readLine()));

        System.out.print("Zahl der Schaetze: ");
        l.setSchaetze(
            Integer.parseInt(System.console().readLine()));

        l.labyrinthErstellen();
        l.spielen();
        ///////////////////////////////////////////////////
        // Behandeln Sie hier die Exceptions die von l.labyrinthErstellen()
        // geworfen werden koennen.
        ///////////////////////////////////////////////////
    }

    /**
     * Darstellung des Labyrinths als String.
     * <p>
     * Stellt Mauern als {@literal "##"}, freie Felder als
     * {@literal "  "}, Gold als {@literal "GG"} und die Spielfigur
     * als {@literal "<>"} dar.
     * </p>
     * @return Mehrzeiliger String
     */
    public String toString() {
        String rv = "";
        for (int z = 0; z < spielfeld.length; ++z) {
            LabyrinthFeld[] zeile = spielfeld[z];
            for (int s = 0; s < zeile.length; ++s) {
                if (s == spielerSpalte
                        && z == spielerZeile) {
                    rv += "<>";
                } else {
                    switch (spielfeld[z][s]) {
                    case Mauer:
                        rv += "##";
                        break;
                    case Frei:
                        rv += "  ";
                        break;
                    case Schatz:
                        rv += "GG";
                        break;
                    }
                }
            }
            rv += "\n";
        }
        return rv;
    }

    /**
     * Bewegt die Spielfigur auf dem Spielfeld.
     * <p>
     * Die Figur wird in die gegebene {@link Richtung} {@code
     * richtung} bewegt, sofern dies eine gueltige Bewegung der Figur
     * darstellt. Dabei gefunden Schaetze werden aus dem Labyrinth
     * entfernt.
     * </p>
     * @param richtung Die gewuenschte Richtung der Spielfigur.
     */
    private void geheNach(Richtung richtung)
        ///////////////////////////////////////////////////
        // Erweitern Sie die Methodensignatur um eine throws-Angabe.
        ///////////////////////////////////////////////////
    {
        int neueZeile = spielerZeile,
            neueSpalte = spielerSpalte;
        switch (richtung) {
        case Norden:
            neueZeile--;
            break;
        case Osten:
            neueSpalte++;
            break;
        case Sueden:
            neueZeile++;
            break;
        case Westen:
            neueSpalte--;
            break;
        }
        if (neueZeile < 0
                || neueZeile >= getHoehe()
                || neueSpalte < 0
                || neueSpalte >= getBreite()) {
            ///////////////////////////////////////////////////
            // Hier soll eine Exception geworfen werden, da die neue Position
            // des Spieler ausserhalb des Spielfeldes liegt.
            ///////////////////////////////////////////////////
        }
        switch (spielfeld[neueZeile][neueSpalte]) {
        case Mauer:
            ///////////////////////////////////////////////////
            // Hier soll eine Exception geworfen werden, da die neue Position
            // des Spieler auf einem Wand-Feld liegt.
            ///////////////////////////////////////////////////
        case Schatz:
            spielfeld[neueZeile][neueSpalte] =
                LabyrinthFeld.Frei;
            schaetze--;
            // Achtung, kein break! (Absichtlich!)
        case Frei:
            spielerZeile = neueZeile;
            spielerSpalte = neueSpalte;
        }
    }

    /**
     * Fuehrt ein Spiel aus.
     * <p>
     * Der Spieler wird wiederholt aufgefordert eine Bewegungsrichtung
     * fuer die Spielfigur anzugegeben. Wurden alle Schaetze im
     * Labyrinth gefunden, hat er das Spiel gewonnen.
     * </p>
     */
    public void spielen() {
        while (getSchaetze() > 0) {
            System.out.println(toString());
            System.out.println("Buchstaben und danach Enter druecken. "
                + "w/a/s/d = Norden/Westen/Sueden/Osten, q = Beenden");
            System.out.print("> ");
            String kommando =
                System.console().readLine();
            switch (kommando) {
            case "w":
                geheNach(Richtung.Norden);
                break;
            case "a":
                geheNach(Richtung.Westen);
                break;
            case "s":
                geheNach(Richtung.Sueden);
                break;
            case "d":
                geheNach(Richtung.Osten);
                break;
            case "q":
                return;
            }
            ///////////////////////////////////////////////////
            // Behandeln Sie hier die Exceptions die von den vier
            // geheNach(...)-Aufrufen geworfen werden koennen.
            ///////////////////////////////////////////////////
        }
        System.out.println("Alle Schaetze gefunden!");
    }
}
