import java.util.ArrayList;
import java.util.Random;

public class LabyrinthGenerator {

    private static class Pair {
        final int row;
        final int col;

        Pair(int row, int col) {
            this.row = row;
            this.col = col;
        }

        Pair move(int i) {
            switch (i) {
            case 0:
                return new Pair(row + 1, col);
            case 1:
                return new Pair(row - 1, col);
            case 2:
                return new Pair(row, col + 1);
            case 3:
                return new Pair(row, col - 1);
            }
            return null;
        }

        @Override
        public String toString() {
            return row + "," + col;
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result + this.col;
            result = prime * result + this.row;
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            Pair other = (Pair) obj;
            if (this.col != other.col)
                return false;
            if (this.row != other.row)
                return false;
            return true;
        }
    }

    /**
     * Erstellt ein Labyrinth.
     * <p>
     * Der Eingang in das Labyrinth ist das zweite Feld in der ersten Zeile,
     * also das Feld {@code l[0][1]}, falls {@code l} das zurueckgegebene Array
     * ist.
     * </p>
     * @param breite Breite des Labyrinths. Schoener fuer ungerade Zahlen.
     * @param hoehe Hoehe des Labyrinths. Schoener fuer ungerade Zahlen.
     * @param anzahlSchaetze Zahl der im Labyrinth platzierten Schaetze.
     * @return Ein zweidimensionales, rechteckiges Array der Felder des
     * Labyrinths.
     * @throws GroesseZuKlein Wird geworfen, falls {@code breite} oder
     * {@code hoehe} zu klein angegeben werden.
     * @throws ZuVieleSchaetze Wird geworfen, falls, relativ zur Groesse des
     * Labyrinths, zu viele Schaetze positioniert werden sollen.
     */
    public static LabyrinthFeld[][] generateLabyrinth(int breite,
        int hoehe,
        int anzahlSchaetze)
        ///////////////////////////////////////////////////
        // Erweitern Sie die Methodensignatur um eine throws-Angabe.
        ///////////////////////////////////////////////////
        {

        if (hoehe < 5) {
            ///////////////////////////////////////////////////
            // Hier soll eine Exception geworfen werden, da die Hoehe des
            // Spielfeldes zu gering ist.
            ///////////////////////////////////////////////////
        }
        if (breite < 5) {
            ///////////////////////////////////////////////////
            // Hier soll eine Exception geworfen werden, da die Breite des
            // Spielfeldes zu gering ist.
            ///////////////////////////////////////////////////
        }
        LabyrinthFeld[][] spielfeld =
            new LabyrinthFeld[hoehe][breite];
        for (int i = 0; i < hoehe; ++i) {
            for (int j = 0; j < breite; ++j) {
                spielfeld[i][j] =
                    LabyrinthFeld.Mauer;
            }
        }
        Random rand = new Random();

        // Eingang
        spielfeld[0][1] = LabyrinthFeld.Frei;

        // Wege bauen
        if (rand.nextBoolean()) {
            buildPathHard(1, 1, spielfeld, rand,
                breite, hoehe);
        } else {
            buildPathEasy(1, 1, spielfeld, rand,
                breite, hoehe);
        }

        // Schaetze platzieren
        int fwidth = (breite - 1) / 2;
        int fheight = (hoehe - 1) / 2;
        if (anzahlSchaetze < 0
            || anzahlSchaetze > fwidth * fheight
                / 4) {
            ///////////////////////////////////////////////////
            // Hier soll eine Exception geworfen werden, da die gewuenschte
            // Zahl der Schaetze zu gross ist.
            ///////////////////////////////////////////////////
        }
        while (anzahlSchaetze > 0) {
            int col =
                rand.nextInt(fwidth) * 2 + 1;
            int row =
                rand.nextInt(fheight) * 2 + 1;
            if (!spielfeld[row][col].equals(LabyrinthFeld.Frei)) {
                continue;
            }
            spielfeld[row][col] =
                LabyrinthFeld.Schatz;
            --anzahlSchaetze;
        }
        return spielfeld;
    }

    private static void buildPathHard(int row,
        int col,
        LabyrinthFeld[][] spielfeld,
        Random rand,
        int breite,
        int hoehe) {
        ArrayList<Pair> todo = new ArrayList<>();
        todo.add(new Pair(row, col));

        breite =
            breite % 2 == 1 ? breite : breite - 1;
        hoehe =
            hoehe % 2 == 1 ? hoehe : hoehe - 1;

        while (!todo.isEmpty()) {
            Pair p =
                todo.remove(rand.nextInt(todo.size()));
            if (spielfeld[p.row][p.col].equals(LabyrinthFeld.Frei)) {
                continue;
            }
            spielfeld[p.row][p.col] =
                LabyrinthFeld.Frei;
            boolean connected = false;
            for (int i = 0; i < 4; ++i) {
                Pair n = p.move(i);
                if (n.row < 0 || n.row >= hoehe
                    || n.col < 0
                    || n.col >= breite) {
                    continue;
                }
                if (spielfeld[n.row][n.col].equals(LabyrinthFeld.Frei)) {
                    connected = true;
                    break;
                }
            }
            if (!connected) {
                for (int i = 0; i < 4; ++i) {
                    Pair t = p.move(i);
                    Pair n = t.move(i);
                    if (n.row < 0
                        || n.row >= hoehe
                        || n.col < 0
                        || n.col >= breite) {
                        continue;
                    }
                    if (spielfeld[n.row][n.col].equals(LabyrinthFeld.Frei)) {
                        spielfeld[t.row][t.col] =
                            LabyrinthFeld.Frei;
                        break;
                    }
                }
            }

            while (true) {
                for (int i = 0; i < 4; ++i) {
                    Pair n2 = p.move(i).move(i);
                    if (n2.row < 0
                        || n2.row >= hoehe
                        || n2.col < 0
                        || n2.col >= breite) {
                        continue;
                    }
                    if (!spielfeld[n2.row][n2.col].equals(LabyrinthFeld.Frei)) {
                        todo.add(n2);
                    }
                }
                int dir = rand.nextInt(4);
                Pair t = p.move(dir);
                Pair n = t.move(dir);
                if (n.row < 0 || n.row >= hoehe
                    || n.col < 0
                    || n.col >= breite) {
                    break;
                }
                if (spielfeld[n.row][n.col].equals(LabyrinthFeld.Frei)) {
                    break;
                }
                spielfeld[t.row][t.col] =
                    LabyrinthFeld.Frei;
                spielfeld[n.row][n.col] =
                    LabyrinthFeld.Frei;
                p = n;
            }

        }
    }

    private static void buildPathEasy(int row,
        int col,
        LabyrinthFeld[][] spielfeld,
        Random rand,
        int breite,
        int hoehe) {
        int[] dirs = {0, 1, 2, 3 };
        int t;

        spielfeld[row][col] = LabyrinthFeld.Frei;

        // permute badly... but simple
        for (int i = 0; i < dirs.length; ++i) {
            int j = rand.nextInt(dirs.length - i);
            t = dirs[i];
            dirs[i] = dirs[j];
            dirs[j] = t;
        }

        for (int richtung : dirs) {
            int trow = row, tcol = col, nrow =
                row, ncol = col;
            switch (richtung) {
            case 0:
                trow--;
                nrow -= 2;
                break;
            case 1:
                tcol++;
                ncol += 2;
                break;
            case 2:
                trow++;
                nrow += 2;
                break;
            case 3:
                tcol--;
                ncol -= 2;
                break;
            }
            if (nrow < 0 || nrow >= hoehe - 1
                || ncol < 0 || ncol >= breite - 1) {
                continue;
            }
            if (spielfeld[trow][tcol].equals(LabyrinthFeld.Frei)) {
                continue;
            }
            if (spielfeld[nrow][ncol].equals(LabyrinthFeld.Mauer)) {
                spielfeld[trow][tcol] =
                    LabyrinthFeld.Frei;
                buildPathEasy(nrow, ncol,
                    spielfeld, rand, breite,
                    hoehe);
            }
        }
    }

}
